;; -*- mode: common-lisp; package: user -*-
;;
;; copyright (c) 2000-2002 Franz Inc, Berkeley, CA
;;
;; The software, data and information contained herein are proprietary
;; to, and comprise valuable trade secrets of, Franz, Inc.  They are
;; given in confidence by Franz, Inc. pursuant to a written license
;; agreement, and may be stored and used only in accordance with the terms
;; of such license.
;;
;; Restricted Rights Legend
;; ------------------------
;; Use, duplication, and disclosure of the software, data and information
;; contained herein by any agency, department or entity of the U.S.
;; Government are subject to restrictions of Restricted Rights for
;; Commercial Software developed at private expense as specified in
;; DOD FAR Supplement 52.227-7013 (c) (1) (ii), as applicable.

;; $Id: rpc-nameserve.cl,v 1.1.4.3 2002/02/07 16:43:02 layer Exp $

(in-package #:user)

(eval-when (compile load eval) (require :aclrpc))

(defpackage #:user 
  (:use #:net.rpc)
  )

(defvar *name-server-port* 7260)
(defvar *name-server-host* (short-site-name))


;;; This is the top-level function in the server image

(defun start-name-server (&optional (host *name-server-host*) 
				    (port *name-server-port*))
  (make-rpc-server 'rpc-datagram-server 
    :limit nil :open :listener
    :local-host host :local-port port
    ))


;;; These are top-level function in the client image(s)

(defun make-name-server (&optional (host *name-server-host*) 
				     (port *name-server-port*))
  (list 
   (setf *name-server-host* host)
   (setf *name-server-port* port)))

(defun make-name-client ()
  (make-rpc-client 'rpc-datagram-port
    :remote-host *name-server-host*
    :remote-port *name-server-port*
    :message-timeout 1
    :verbose nil
    ))

(defun name-register (name host port)
  (let ((rport (make-name-client)))
    (unwind-protect
	(dotimes (i 5 (values nil :timeout))
	  (case (ignore-errors (rpc-invoke rport 'do-name-register (list name host port)
					   :wait t :error-p nil))
	    (:done (return :done))
	    (:duplicate (return (values nil :duplicate)))
	    ))
      (rpc-close rport :stop t)
      (rpc-close rport :stop t :abort t))))
	    


(defun name-lookup (name)
  (let ((port (make-name-client)))
    (unwind-protect
	(dotimes (i 5 (values nil :timeout))
	  (let* ((m (rpc-invoke port 'do-name-lookup (list name)
				:wait nil))
		 s)
	    (setf s (multiple-value-list (rpc-wait port m)))
	    (or (car s)
		(return (values-list s)))
	    (case (setf s (rpc-query port m))
	      (:rms-done (return (multiple-value-list (rpc-get port m))))
	      ((:rms-error :rms-throw)
	       (return (values nil s (rpc-get port m :error-p nil)))))))
      (rpc-close port :stop t)
      (rpc-close port :stop t :abort t))))



;;; This is the code that does the work in the server image

(defvar *name-registry* (make-hash-table :test #'equalp))

(defun do-name-register (name host port)
  (if* (gethash name *name-registry*)
       then :duplicate
       else
       (setf (gethash name *name-registry*)
	     (list host port))
       :done))

(defun do-name-lookup (name) 
  (values-list (gethash name *name-registry*)))
  



				    


