;; copyright (c) 1985 Franz Inc, Alameda, Ca.
;; copyright (c) 1986-2001 Franz Inc, Berkeley, CA  - All rights reserved.
;;
;; The software, data and information contained herein are proprietary
;; to, and comprise valuable trade secrets of, Franz, Inc.  They are
;; given in confidence by Franz, Inc. pursuant to a written license
;; agreement, and may be stored and used only in accordance with the terms
;; of such license.
;;
;; Restricted Rights Legend
;; ------------------------
;; Use, duplication, and disclosure of the software, data and information
;; contained herein by any agency, department or entity of the U.S.
;; Government are subject to restrictions of Restricted Rights for
;; Commercial Software developed at private expense as specified in
;; DOD FAR Supplement 52.227-7013 (c) (1) (ii), as applicable.

;; Bidirectional (dual channel) rot13 encapsulator

(in-package :user)

(eval-when (compile)
  (require :iodefs))


(def-stream-class rot13-bidirectional-stream
    (string-input-simple-stream string-output-simple-stream)
  ())

(defmethod device-open ((stream rot13-bidirectional-stream) options)
  (let ((base-stream (getf options :base-stream)))
    (if base-stream
	(with-stream-class (rot13-bidirectional-stream stream)
	  (when (input-stream-p base-stream)
	    (unless (sm excl::buffer stream)
	      (setf (sm excl::buffer stream)
		(make-string (device-buffer-length stream))))
	    (setf (sm excl::input-handle stream) base-stream)
	    (add-stream-instance-flags stream :string :input :simple))
	  (when (output-stream-p base-stream)
	    (unless (sm excl::out-buffer stream)
	      (setf (sm excl::out-buffer stream)
		(make-string (device-buffer-length stream))))
	    (setf (sm excl::max-out-pos stream) (length (sm excl::out-buffer stream)))
	    (setf (sm excl::output-handle stream) base-stream)
	    (add-stream-instance-flags stream :string :output :simple))
	  (install-string-character-strategy stream)
	  (when (interactive-stream-p base-stream)
	    (setf (interactive-stream-p stream) t))
	  t)
      (progn (error "rot13-bidirectional-stream needs a :base-stream")
	     nil))))

(defmethod device-extend ((stream rot13-bidirectional-stream) need action)
  (with-stream-class (rot13-bidirectional-stream stream)
    (let ((base-stream (or (sm excl::input-handle stream) ; One of these should
			   (sm excl::output-handle stream)))) ; be non-nil
      (ecase action
	(:input-check
	 (let* ((buffer (sm excl::buffer stream))
		(len (length buffer)))
	   (setf (sm excl::buffpos stream) 0)
	   (when (read-no-hang-p base-stream)
	     (let ((res (read-vector buffer base-stream :start 0 :end len)))
	       (when (<= res 0)
		 (setf (sm excl::buffer-ptr stream) 0)
		 (return-from device-extend nil))
	       (dotimes (i res)
		 (setf (char buffer i)
		   (rotate-char (char buffer i))))
	       (setf (sm excl::buffer-ptr stream) res)
	       t))))
	(:input
	 (let* ((buffer (sm excl::buffer stream))
		(len (length buffer))
		(cur 0)
		(saw-newline nil)
		(interactive (interactive-stream-p base-stream)))
	   (setf (sm excl::buffpos stream) 0)
	   (loop
	     (let ((res (read-vector buffer base-stream :start cur :end len)))
	       (when (<= res 0)
		 (setf (sm excl::buffer-ptr stream) cur)
		 (return-from device-extend (> cur 0)))
	       (do ((i cur (1+ i))
		    (j 0 (1+ j)))
		   ((eq j res))
		 (let ((char (char buffer i)))
		   ;; This hack is to allow *terminal-io* to do something
		   ;; on a per-line basis
		   (when (and (eql #\Newline char)
			      interactive)
		     (setq saw-newline t))
		   (setf (char buffer i)
		     (rotate-char char))))
	       (incf cur res)
	       (when (or (>= cur need) saw-newline)
		 (setf (sm excl::buffer-ptr stream) cur)
		 (return t))))))
	(:output-check
	 (write-no-hang-p base-stream))
	(:output
	 (let ((res (device-write stream (sm excl::out-buffer stream)
				  0 (sm excl::out-pos stream) t)))
	   (when (> res 0)
	     (setf (sm excl::out-pos stream) 0)
	     t)))))))



(defun rotate-char (char)
  (unless (alpha-char-p char)
    (return-from rotate-char char))
  (let ((code (char-code char)))
    (cond ((<= #.(char-code #\A) code #.(char-code #\Z))
	   (code-char
	    (+ #.(char-code #\A)
	       (mod (+ 13 (- code #.(char-code #\A))) 26))))
	  (t
	   ;; assume ascii; must be lowercase letter
	   (code-char
	    (+ #.(char-code #\a)
	       (mod (+ 13 (- code #.(char-code #\a))) 26)))))))

(defmethod device-file-length ((stream rot13-bidirectional-stream))
  nil)

(defmethod device-file-position ((stream rot13-bidirectional-stream))
  nil)

(defmethod (setf device-file-position) (new-value (stream rot13-bidirectional-stream))
  (declare (ignore new-value))
  nil)

(defmethod device-write ((stream rot13-bidirectional-stream) buffer start end blocking)
  (declare (ignore blocking))
  (when (and (null buffer) (not (eq start end)))
    (with-stream-class (rot13-bidirectional-stream)
      (setq buffer (sm excl::out-buffer stream))))
  (with-stream-class (rot13-bidirectional-stream stream)
    (let ((base-stream (sm excl::output-handle stream)))
      (if (and base-stream (> end start))
	  (let ((cur start))
	    (do ((i start (1+ i)))
		((>= i end))
	      (setf (char buffer i)
		(rotate-char (char buffer i))))
	    (loop
	      (let ((res (write-vector buffer base-stream :start cur :end end)))
		(cond ((< res 0)
		       (return-from device-write res))
		      ((= res 0)
		       (return-from device-write cur)))
		(incf cur res)
		(when (>= cur end)
		  (return-from device-write cur)))))
	0))))
