;;;   -*- Mode: LISP; Syntax: Common-lisp; Base: 10; Lowercase:T -*-

;; $Id: extloop.cl,v 1.8 1997/12/17 17:48:51 layer Exp $

(provide :extloop)

(in-package :excl)

(eval-when (compile eval load) (require :loop))

#+Cloe-Runtime					;Don't ask.
(car (push "%Z% %M% %I% %E% %U%" system::*module-identifications*))


;;; The following code could be used to set up the SYMBOLICS-LOOP package
;;; as it is expected to be.  At Symbolics, in both Genera and CLOE, the
;;; package setup is done elsewhere.   


#-Symbolics
(eval-when (compile eval load)
  (unless (find-package 'symbolics-loop)
    (make-package 'symbolics-loop :use nil)))

#-Symbolics
(import 'excl::loop-finish (find-package 'symbolics-loop))

#-Symbolics
(export '(symbolics-loop::loop
		   symbolics-loop::loop-finish
		   symbolics-loop::define-loop-iteration-path
		   symbolics-loop::define-loop-sequence-path
		   )
	(find-package 'symbolics-loop))



;;;This is our typical "extensible" universe, which should be a proper superset of the ansi universe.
(defvar *loop-default-universe*
    (make-ansi-loop-universe t))





(defmacro symbolics-loop:define-loop-iteration-path (path-name function
						     &key alternate-names preposition-groups
						     inclusive-permitted user-data (loop-universe '*loop-default-universe*))
  `(eval-when (eval compile load)
     (add-loop-path '(,path-name ,@alternate-names) ,function ,loop-universe
		    :preposition-groups ',preposition-groups
		    :inclusive-permitted ',inclusive-permitted
		    :user-data ',user-data)))


(defmacro symbolics-loop:define-loop-sequence-path (path-name-or-names fetch-function size-function
						    &optional sequence-type element-type)
  "Defines a sequence iteration path.  PATH-NAME-OR-NAMES is either an
atomic path name or a list of path names.  FETCHFUN is a function of
two arguments, the sequence and the index of the item to be fetched.
Indexing is assumed to be zero-origined.  SIZEFUN is a function of
one argument, the sequence; it should return the number of elements in
the sequence.  SEQUENCE-TYPE is the name of the data-type of the
sequence, and ELEMENT-TYPE is the name of the data-type of the elements
of the sequence."
  `(eval-when (eval compile load)
     (add-loop-path ',path-name-or-names 'loop-sequence-elements-path *loop-default-universe*
		    :preposition-groups '((:of :in) (:from :downfrom :upfrom) (:to :downto :below :above) (:by))
		    :inclusive-permitted nil
		    :user-data '(:fetch-function ,fetch-function
				 :size-function ,size-function
				 :sequence-type ,sequence-type
				 :element-type ,element-type))))


(defmacro symbolics-loop:loop (&environment env &rest keywords-and-forms)
  #+Genera (declare (compiler:do-not-record-macroexpansions)
		    (zwei:indentation . zwei:indent-loop))
  (loop-standard-expansion keywords-and-forms env *loop-default-universe*))
