;; copyright (c) 1986-2002 Franz Inc, Berkeley, CA  - All rights reserved.
;;
;; Allegro CL IPC interface, for versions 4.2 and later.
;;
;; This code is provided to licensed users of Allegro CL as an example
;; and for restricted reuse.


;;;;; Hacked to support the LEP protocol
;;;; This leaves some room for improvement
;;;; Changes:
;;;; 1. Support for multiple servers listening on different connections.
;;;; 2. The default behaviour is that the server reads a protocol name
;;;;    that specifies the type of the server process to create.
;;;; Currently, supported protocols are:
;;;;    a. :listener - RUN a listener
;;;;    b. :lep - start up an LEP server

(eval-when (compile eval load)
  (if (excl::resolve-symbol-in-package :rcsnote :sys)
      (push :has-rcsnote *features*))

  #-(or little-endian big-endian)
  (pushnew
   (let ((x '#(1)))
     (if (not (= 0 (sys::memref x #.(sys::mdparam 'comp::md-svector-data0-adj)
				0 :unsigned-byte)))
	 :little-endian
       :big-endian))
   *features*)
  )

#+has-rcsnote
(si::rcsnote "ipc" "$Revision: 1.97.4.3.6.1 $")

(eval-when (compile eval load)
  (excl::require :process)
  (excl::require :foreign)
  (excl::provide :ipc)
  #+(version>= 5 0) (require :sock)
  )

(defpackage :ipc
  (:use :common-lisp :excl :ff :mp)
  (:export #:start-lisp-listener-daemon #:open-network-stream
	   #:*inet-port-min* #:*inet-port-max* #:*inet-port-used*
	   #:start-process-for-network-stream))

(in-package :ipc)



;; So we don't get warnings in sys:compile-cl:
(declaim (:fbound lisp_htonl lisp_ntohl accept unix-listen
		  lisp_htons bind unix-close perror socket))

(let ((p *package*))
  (setf (package-lock            p) t
	(package-definition-lock p) t))

(eval-when (compile eval load)
  (excl::require :defctype))

(defvar *unix-domain* nil
  ;; If non-nil then use a UNIX domain socket, otherwise use an internet
  ;; domain port (see *inet-port* variable).
  )

(defvar *socket-pathname* nil
  ;; When the UNIX domain is used, then if non-nil this is the pathname of
  ;; the socket file to use for the communication between GNU Emacs and
  ;; Allegro CL.
  )

(defparameter *inet-port-min* 1025
  "The smallest internet service port number on which Lisp listens for
connections.")

(defparameter *inet-port-max* 2025
  "The largest internet service port number on which Lisp listens for
connections.")

(defvar *inet-port-used* 0
  "Port actually in use for lisp listeners.")

(defvar *inet-listener-password* 0
  ;; A magic number that mnust be supplied to verify that someone trying to
  ;; connect to our listener is really the person who started the lisp
  )

(defvar *ipc-version* 1
  ;; This indicates the version of the IPC protocol
  )

(defconstant *af-unix* 1
  ;; The AF_UNIX constant from /usr/include/sys/socket.h.
  )

(defconstant *af-inet* 2
  ;; The AF_INET constant from /usr/include/sys/socket.h.
  )

(defconstant *sock-stream* #+svr4 2 #-svr4 1
  ;; The SOCK_STREAM constant from /usr/include/sys/socket.h.
  )

;; from <netinet/in.h>
(defcstruct sockaddr-in
  (family :unsigned-short)		; short sin_family
  (port :unsigned-short)		; u_short sin_port
  (addr :unsigned-int)			; struct in_addr sin_addr
  (zero 8 :char)			; char sin_zero[8]
  )





;; from <sys/un.h>

(defcstruct sockaddr-un
  (family :unsigned-short)
  (path 109 :char))







(defcstruct (servent :malloc)		; Returned by getservent
  (name * :char)
  (aliases * * :char)
  (port :signed-int)
  (proto * :char))





(defcstruct timeval
  (sec :int)
  (usec :int))








(defcstruct unsigned-int
  (unsigned-int :unsigned-int))







#+mswindows
(defcstruct (w-s-a-data :malloc)
  (w-version :short)
  (w-high-version :short)
  (description 257 :char)
  (system-status 129 :char)
  (max-sockets :unsigned-short)
  (max-udp-dg :unsigned-short)
  (vendor-info * :char))






(defvar .lisp-listener-daemon-ff-loaded. nil)
(defvar .lisp-listener-daemon. nil)

(defparameter .needed-funcs.
    '("socket" "bind" "listen" "accept" "getsockname" "connect"))

#+ignore
(defun entry-point-exists-p (string)
  (get-entry-point string))

#+ignore
(defvar .junk-name. (make-array 1))
#+ignore
(defvar .junk-address. (make-array 1 :element-type '(unsigned-byte 32)))

#+ignore
(defun entry-point-exists-p (string)
  (setf (aref .junk-name. 0) string)
  (setf (aref .junk-address. 0) 0)
  (= 0 (get-entry-points .junk-name. .junk-address.)))

#+mswindows
(defvar *winsock-version-data* nil)


#-(version>= 5 0) ;;
(defun convert-foreign-name (name &key (language :c))
  (convert-to-lang name :language language))





(eval-when (load eval)
  (unless .lisp-listener-daemon-ff-loaded.
    (excl::machine-case :host
      ((:rs6000))
      ((:msx86 :msdec)
       (let ((windir (sys::getenv "WINDIR")))
	 (unless windir (error "WINDIR not defined in environment"))
	 (if* (probe-file (format nil "~a\\system32\\" windir))
	    then ;; NT
		 (load (format nil "~a\\system32\\wsock32.dll" windir))
		 (load (format nil "~a\\system32\\msvcrt20.dll" windir))
	    else ;; 95
		 (load (format nil "~a\\system\\wsock32.dll" windir))
		 (load (format nil "~a\\system\\msvcrt20.dll" windir)))))
      (t (unless (dolist (name .needed-funcs. t)
		   (if (not (get-entry-point
			     (convert-foreign-name name)))
		       (return nil)))
	   (princ ";  Loading networking routines from C library...")
	   (force-output)
	   (unless (load ""
			 :verbose nil
			 :unreferenced-lib-names
			 (mapcar #'convert-foreign-name .needed-funcs.)
			 #+ignore :system-libraries
			 #+ignore '("bsd")
			 )
	     (error "foreign load failed"))
	   (princ "done")
	   (terpri)
	   (force-output))))

    (setq .lisp-listener-daemon-ff-loaded. t)
    (defforeign-list `((w-s-a-startup :entry-point ,(list "WSAStartup")
				      :strings-convert t)
		       (socket :entry-point ,(list "socket")
			       :strings-convert t)
		       (bind :entry-point ,(list "bind")
			     :strings-convert t)
		       (accept :entry-point ,(list "accept")
			       :strings-convert t)
		       (getsockname :entry-point ,(list "getsockname")
				    :strings-convert t)
		       (select :entry-point ,(list "select")
			       :strings-convert t)
		       (perror :entry-point ,(list "perror")
			       :strings-convert t)
		       (unix-listen :entry-point ,(list "listen")
				    :strings-convert t)
		       (unix-close
			:entry-point ,(list
				       #-mswindows "close"
				       #+mswindows "closesocket")
			:strings-convert t))
	:print nil)

    #+mswindows
    (let ((errcode nil))
      (setq *winsock-version-data* (make-cstruct 'w-s-a-data))
      (setq errcode (w-s-a-startup (+ 1 (ash 1 8)) *winsock-version-data*))
      (unless (zerop errcode)
	(error "WSAStartup failed return code ~s." errcode))
      (let* ((v (w-s-a-data-w-version *winsock-version-data*))
	     (major (ash v -8))
	     (minor (logand v #xff)))
	(if* (and (= 1 major) (= 1 minor))
	 then (format t	"; Initialization of wsock32.dll v1.1 successful~%")
	   else (error "WSAStartup failed."))))
    ))



(if* (get-entry-point (convert-foreign-name "lisp_htons"))
   then ;; Allegro CL 3.1 or later...
	(defforeign-list `((lisp_htons :entry-point ,(list "lisp_htons")
				       :strings-convert t)
			   (lisp_htonl :entry-point ,(list "lisp_htonl")
				       :strings-convert t)
			   (lisp_ntohs :entry-point ,(list "lisp_ntohs")
				       :strings-convert t)
			   (lisp_ntohl :entry-point ,(list "lisp_ntohl")
				       :strings-convert t))
			 :print nil)
   else ;; pre-3.1 Allegro CL.  Do it the hard way...
	#+little-endian
	(progn
	  (setf (symbol-function 'lisp_htons)
	    #'(lambda (x)
		(logior (ash (logand x #x00ff) 8)
			(ash (logand x #xff00) -8))))
	  (setf (symbol-function 'lisp_ntohs)
	    #'(lambda (x) (lisp_htons x)))

	  (setf (symbol-function 'lisp_htonl)
	    #'(lambda (x)
		(logior (ash (logand x #x000000ff)  24)
			(ash (logand x #x0000ff00)   8)
			(ash (logand x #x00ff0000)  -8)
			(ash (logand x #xff000000) -24))))
	  (setf (symbol-function 'lisp_ntohl)
	    #'(lambda (x) (lisp_htonl x))))

	#+big-endian
	(progn			;Each used to be #'(lambda (x) x) -smh
	  (setf (symbol-function 'lisp_htons) #'identity)
	  (setf (symbol-function 'lisp_htonl) #'identity)
	  (setf (symbol-function 'lisp_ntohs) #'identity)
	  (setf (symbol-function 'lisp_ntohl) #'identity)))



(defun start-lisp-listener-daemon (&rest args)
  "Starts a daemon process which listens to a socket for attempts to
connect, and starts a lisp listener for each connection.  If the Lisp
listener ever completes, the daemon makes sure its connection is closed."
  (cond ((or (null .lisp-listener-daemon.)
	     (not (member .lisp-listener-daemon. sys::*all-processes*
			  :test #'eq)))
	 (setq .lisp-listener-daemon.
	   (apply #'process-run-function "Listener Socket Daemon"
		  'lisp-server-socket-daemon
		  args))
	 (setf (mp::process-interruptible-p .lisp-listener-daemon.) nil)
	 (setf (mp::process-keeps-lisp-alive-p .lisp-listener-daemon.) nil); bug11128
	 )
	(t (warn "the listener daemon is already running"))))

(defun kill-lisp-listener-daemon ()
  (if* (or (null .lisp-listener-daemon.)
	   (not (member .lisp-listener-daemon. sys::*all-processes*
			:test #'eq))
	   (progn
	     (mp:process-kill .lisp-listener-daemon.)
	     ;; this makes sure the kill is allowed to complete:
	     (mp:process-allow-schedule)
	     t))
     then (setq .lisp-listener-daemon. nil)
	  t
     else (error "couldn't kill the listener daemon")))

(defun start-lisp-server-process (&rest args
				  &key (server-name
					"Listener Socket Daemon")
				  &allow-other-keys)
  ;;(remf args :server-name)
  (let ((proc
	 (apply #'process-run-function server-name
		'lisp-server-socket-daemon
		args)))
    (setf (mp::process-interruptible-p proc) nil)
    (setf (mp::process-keeps-lisp-alive-p proc) nil); bug11128
    proc))

(defun lisp-server-socket-daemon (&rest args
				  &key
				  (use-lep nil)
				  (startup-stream-process
				   (if use-lep
				       #'new-default-stream-startup
				     #'default-stream-startup))
				  (server-startup-announce-function
				   #'announce-server-startup-to-lep)
				  ((:unix-domain *unix-domain*) *unix-domain*)
				  (inet-port 0 inet-port-p)
				  ((:inet-port-min *inet-port-min*)
				   (if inet-port-p inet-port *inet-port-min*))
				  ((:inet-port-max *inet-port-max*)
				   (if inet-port-p inet-port *inet-port-max*))
				  ((:socket-pathname *socket-pathname*)
				   *socket-pathname*)
				  &allow-other-keys
;;; We should decide this if we
;;; implement background streams
;;; then it is no good just outputting to random locations.
				  &aux
				  (*standard-output* *initial-terminal-io*)
				  (*debug-io* *initial-terminal-io*)
				  (*terminal-io* *initial-terminal-io*)
				  (*error-output* *initial-terminal-io*))

  ;; use-lep should default to Nil and the LEP should start things up with T
  ;; In this way a new lisp can work with the old Emacs interface

  (block bad-news
    (let (listen-socket-fd
	  client-address
	  (listen-sockaddr
	   (if *unix-domain*
	       (make-cstruct 'sockaddr-un)
	     (let ((sin (make-cstruct 'sockaddr-in)))
	       (excl::memset sin 0 (ff::cstruct-length 'sockaddr-in))
	       sin)))
	  (timeval (make-cstruct 'timeval))
	  (int (make-cstruct 'unsigned-int))
	  stream
	  fd)

      (unless *socket-pathname*
	(setq *socket-pathname* (sys:make-temp-file-name "GnuToAcl")))

      (setf (timeval-sec timeval) 0
	    (timeval-usec timeval) 0)
      (unwind-protect
	  (progn
	    (if *unix-domain* (errorset (delete-file *socket-pathname*)))
	    (setq listen-socket-fd (socket
				    (if *unix-domain* *af-unix* *af-inet*)
				    *sock-stream*
				    0))
	    (when (< listen-socket-fd 0)
	      (perror "socket")
	      (setq listen-socket-fd nil)
	      (return-from bad-news))

	    (mp:waiting-for-input-available (listen-socket-fd)
	     (if* *unix-domain*
		then (setf (sockaddr-un-family listen-sockaddr) *af-unix*)
		     ;; Set pathname.
		     (dotimes (i (length *socket-pathname*)
				(setf (sockaddr-un-path listen-sockaddr i) 0))
		       (setf (sockaddr-un-path listen-sockaddr i)
			 (char-int (elt *socket-pathname* i))))
		     (unless (zerop (bind listen-socket-fd
					  listen-sockaddr
					  (+ (length *socket-pathname*) 2)))
		       (perror "bind")
		       (return-from bad-news))
		else (setf (sockaddr-in-family listen-sockaddr) *af-inet*)
		     (do ((port *inet-port-min* (1+ port)))
			 ((progn
			    (setf (sockaddr-in-port listen-sockaddr)
			      (lisp_htons (setq *inet-port-used* port)))
			    (zerop (bind listen-socket-fd
					 listen-sockaddr
					 (ff::cstruct-length 'sockaddr-in)))))
		       (if* (= port *inet-port-max*)
			  then (perror "bind")
			       (return-from bad-news))))

	     (finish-output)
	     (unless (zerop (unix-listen listen-socket-fd 5))
	       (perror "listen")
	       (return-from bad-news))

	     (funcall server-startup-announce-function)

	     (loop
	       (process-wait "waiting for a connection"
			     #'excl::read-no-hang-p
			     listen-socket-fd)
	       (setf (unsigned-int-unsigned-int int)
		 (if *unix-domain*
		     (ff::cstruct-length 'sockaddr-un)
		   (ff::cstruct-length 'sockaddr-in)))
	       ;; There are some potential propblems here.  The socket really should
	       ;; be set non-blocking, because an attempt to connect could time out
	       ;; or be aborted between the time the process-wait function completes
	       ;; and the time accept is called.  In this case the call would block.
	       (setq fd (accept listen-socket-fd listen-sockaddr int))
	       (setq client-address
		 (if *unix-domain*
		     (let ((s (make-string :length int)))
		       (loop for i below int
			   do (setf (schar s i)
				(sys::fixnum-to-char (sockaddr-un-path listen-sockaddr i))))
		       s)
		   (format-in-addr nil (sockaddr-in-addr listen-sockaddr))))
	       (finish-output *standard-output*)
	       (finish-output *error-output*)
	       (when (< fd 0)
		 (perror "accept")
		 (return-from bad-news))

	       (setq stream
		 (make-ipc-terminal-stream fd
					   :pretty-id (format nil "from ~a" client-address)))
	       ;; the first thing that comes over the stream is the name of the
	       ;; emacs buffer which was created--we name the process the same.
	       ;; For internet sockets, the next thing is the password.
	       (apply startup-stream-process stream
		      :fd fd
		      :listen-sockaddr listen-sockaddr
		      args))))
	(when listen-socket-fd
	  (unix-close listen-socket-fd)
	  (setq .lisp-listener-daemon. nil)))))
  (error "couldn't start listener daemon"))

;; This is the default for back compatibility.
(defun announce-server-startup-to-lep ()
  ;; We could use *standard-output* itself here, but the startup is
  ;; asynchronous with the initial lisp listener startup, and perturbing
  ;; the stream output column causes unpredictable indenting during
  ;; startup.  So instead we use a distinct stream that shares
  ;; *standard-output*'s fd.
  (let ((s (make-ipc-terminal-stream (stream-output-fn *standard-output*))))
    (write-string
     (format nil "~d ~d ~a ~s ~d"
	     ;; ^Aport password case-mode socket-file ipc-version^A
	     (if* *unix-domain*
		then 0
		else *inet-port-used*)
	     (setq *inet-listener-password*
	       (random 1000000 (make-random-state t)))
	     (case excl::*current-case-mode*
	       ((:case-insensitive-upper
		 :case-sensitive-upper) ":upper")
	       ((:case-insensitive-lower
		 :case-sensitive-lower) ":lower"))
	     (if* *unix-domain*
		then *socket-pathname*
		else nil)
	     *ipc-version*)
     s)
    (finish-output s)))

;;;; The first thing that comes over is the name of the protocol to use.
;;;; The old scheme is the listener protocol which means fire up a listener.
;;;; The LEP uses the :lep protocol

(defun new-default-stream-startup (stream &rest args)
  (declare (ignore function))
  (apply #'start-process-for-network-stream stream (read stream) args))

(defgeneric start-process-for-network-stream (stream protocol &rest arguments)
	    (:documentation "Create the process to serve the stream"))

(defmethod start-process-for-network-stream (stream (protocol (eql :listener)) &rest args)
  (apply #'default-stream-startup stream args))

(defmethod start-process-for-network-stream (stream protocol &rest args)
  (declare (ignore args))
  (warn "Tried to start unknown protocol ~S for stream ~S" protocol stream))

(defun default-stream-startup (stream
			       &key (function 'lisp-listener-with-stream-as-terminal-io)
				    listen-sockaddr fd
			       &allow-other-keys)
  (let ((proc-name (read stream)))
    (unless *unix-domain*
      (let ((password (read stream)))
	(unless (and (numberp password)
		     (= password *inet-listener-password*))
	  (format t ";; access denied for host ")
	  (format-in-addr t (lisp_ntohl (sockaddr-in-addr listen-sockaddr)))
	  (format t ", password ~s~%" password)
	  (refuse-connection fd)
	  (return-from default-stream-startup nil))))
    (process-run-function proc-name function stream)))

(defun refuse-connection (fd &aux s)
  (setq s (make-ipc-terminal-stream fd))
  (format s "connection refused.~%")
  (force-output s)
  (excl::clear-output-1 s)
  (close s))

(defun lisp-listener-with-stream-as-terminal-io (s)
  (unwind-protect
      (tpl:start-interactive-top-level
       s 'tpl:top-level-read-eval-print-loop nil)
    ;; This next crock is to prevent the force-output done by close from
    ;; signalling an error if there are characters buffered to the output
    ;; stream, which there will be if the remote client closed the connection.
    ;; This should be changed to a clear-output once that works on a buffered
    ;; terminal stream.
    (excl::clear-output-1 s)
    (close s)))


(defun open-network-stream (&key host port socket-file class element-type)
  "Open a stream to a port, which is a TCP/IP communication channel.  There
are two types of ports supported, UNIX and INTERNET domain.  The domain is
chosen based on the keyword arguments supplied:
For internet domain:
 HOST is the string host name or an integer internet address.
 PORT is the string service name or a port number.
For Unix domain:
 SOCKET-FILE is the string pathname of the socket.
ELEMENT-TYPE and CLASS are arguments for stream creation, as for OPEN."
  (when (or (and host port socket-file)
	    (and (or (null host) (null port))
		 (null socket-file)))
    (error "Must either supply HOST and PORT *or* SOCKET-FILE keywords."))
  (if* socket-file
     then ;; UNIX domain
	  (let ((server (make-cstruct 'sockaddr-un))
		socket-fd)
	    (setf (sockaddr-un-family server) *af-unix*)
	    (dotimes (i (length socket-file)
		       (setf (sockaddr-un-path server i) 0))
	      (setf (sockaddr-un-path server i)
		(char-int (elt socket-file i))))
	    (setq socket-fd (socket *af-unix* *sock-stream* 0))
	    (unless (eq t (excl::filesys-connect socket-fd server (+ 2 (length socket-file))))
	      (perror (format nil "connect failed to ~s" socket-file)))
	    (make-ipc-terminal-stream socket-fd
				      :class class
				      :element-type element-type
				      :pretty-id (format nil "to ~a" socket-file)))
     else ;; INTERNET domain
	  (let (sock server hostaddress)
	    ;; Open a socket
	    (when (< (setf sock (socket *af-inet* *sock-stream* 0)) 0)
	      (error "couldn't open socket"))
	    (excl::filesys-set-ndelay sock t)
	    ;; construct a socket address
	    (setf server (make-cstruct 'sockaddr-in))
	    (excl::memset server 0 (ff::cstruct-length 'sockaddr-in))
	    (if* (integerp host)
	       then (setq hostaddress (lisp_htonl host))
	     elseif (stringp host)
	       then (setq hostaddress (lisp_htonl
				       (socket::lookup-hostname host)))
	       else (error "HOST not a string or integer internet address: ~s"
			   host))
	    (setf (sockaddr-in-addr server) hostaddress)

	    (setf (sockaddr-in-port server)
	      (if* (integerp port)
		 then (lisp_htons port)
	       elseif (stringp port)
		 then (let ((serv (excl::getservbyname port "tcp")))
			(if* (= 0 serv)
			   then (error "Unknown service name: ~s" port))
			(servent-port serv))
		 else (error "PORT not a string or integer internet service: ~s" port)))

	    (setf (sockaddr-in-family server) *af-inet*)
	    ;; open the connection
	    (when (let ((len (ff::cstruct-length 'sockaddr-in))
			(ofd (- -1 sock)))
		    ;; Fix me -- mp hazard for errno.
		    (mp:waiting-for-input-available ;actually wait for output
		     (ofd)
		     (loop as ret = (excl::filesys-connect sock server len)
			 when (or (eq ret t)
				  (eq ret excl::*error-code-eisconn*))
			 return nil
			 while (or (eq ret excl::*error-code-einprogress*)
				   (eq ret excl::*error-code-ealready*))
			 do (process-wait "waiting for connect to complete"
					  #'excl::read-no-hang-p
					  ofd)
			 finally (return t))))
	      (perror "Connect failure")
	      (unix-close sock)
	      (error "couldn't connect to socket"))
	    ;; build and return the stream
	    (make-ipc-terminal-stream sock
				      :class class
				      :element-type element-type
				      :pretty-id (format nil "to [~a/~a]" host port)))))



(eval-when (compile eval)
  #+mswindows
  (defmacro displace-fd (fd)
    (let ((gfd (gensym))) `(let ((,gfd ,fd)) (+ ,gfd 5000))))
  #-mswindows
  (defmacro displace-fd (fd) fd)
  )

(defun make-ipc-terminal-stream (fd &key pretty-id class element-type)
  (let ((s (make-instance (or class
			      'excl::bidirectional-terminal-stream)
	     :element-type (or element-type 'character)
	     :fn-in  fd ;;(displace-fd fd)
	     :fn-out fd ;;(displace-fd fd)
	     )))
    (when pretty-id
      (setf (getf (excl::stream-property-list s) :pretty-socket-id)
	pretty-id))
    s))

(defun format-in-addr (stm addr)	; assumes host byte order
  (format stm "[~11,1,1,'.<~2,'0D~;~2,'0D~;~2,'0D~;~2,'0D~>]"
	  (ldb (byte 8 24) addr)
	  (ldb (byte 8 16) addr)
	  (ldb (byte 8  8) addr)
	  (ldb (byte 8  0) addr)))


;;; Allegro Common Lisp as the client side of `finger'.
;;; This is a simple demo of how easily a lisp program can become the
;;; client of a client/server connection.

#+example
(defun finger (host &optional (name ""))
  (with-open-stream (s (ipc:open-network-stream :host host
						:port "finger"))
    ;; The first interaction of the finger protocol is to send the
    ;; name in which we are interested.  The empty string will
    ;; report on all current users.
    (format s "~a~%" name)
    ;; Then we read back multiple lines of information.
    (do ((line (read-line s nil nil) (read-line s nil nil)))
	((null line) (values))
      ;; The returned data contains extra #\return characters
      ;; which we might not want to print.
      (write-line (remove #\return line)))))

;; This function starts a lisp-listener server to which telnet can connect.
;; Beware, telnet control data can confuse the listener.
#+example
(defun start-telnet-server ()
  (let (inet-port)
    (ipc::start-lisp-server-process
     :server-name "Telnet Server"
     :startup-stream-process #'(lambda (stm &rest args)
				 (declare (ignore args))
				 (let ((client-address
					(getf (excl::stream-property-list stm)
					      :pretty-socket-id)))
				   (format t "~&;Accepting telnet connection ~a/~a~%"
					   client-address inet-port)
				   (mp:process-run-function
				    (format nil "Telnet Client~@[ ~a~]"
					    client-address)
				    #'ipc::lisp-listener-with-stream-as-terminal-io
				    stm)))
     :server-startup-announce-function #'(lambda ()
					   (setq inet-port ipc:*inet-port-used*)
					   (setf (mp:process-name mp:*current-process*)
					     (format nil "Telnet Server port ~a" inet-port))
					   (format t "~&;Telnet server started on port ~a~%"
						   inet-port)))))

;;; Another client example.
;;; This returns what get-universal-time would return on a remote host.

#+example
(defun get-universal-time-remote (host)
  (with-open-stream (s (ipc:open-network-stream :host host
						:port "time"))
    (let ((tod (loop as c = (read-char s nil nil)
		   while c
		   collect c)))
      (when (= 4 (length tod))
	(let ((vec (make-array 4 :element-type 'character
			       :initial-contents tod)))
	  (logand #xffffffff
		  (ipc::lisp_ntohl
		   (sys:memref vec 0 2 :unsigned-long32))))))))
