# $Id: runlisp.sh,v 1.8.22.1 2002/02/08 19:09:15 layer Exp $
#
# A portable way to run lisp from a makefile.  Needed because lisp.exe
# on Windows can't read from `stdin' and you can't redirect the output
# either.
#
# On Windows, use of this script relies on the GNU-Win32 tools, which are
# a port of the popular GNU development tools to Windows NT and 95.
# See http://www.cygnus.com/misc/gnu-win32/ for more information.
#
# Usage: runlisp [-e] [-i] [-t title] [-o output_file] [-f file | form] ...
# run
#		-e	 ignore the exit status (for printing a message about the errors)
#		-i	 run in `interactive' mode (ie, lisp doesn't exit if an error
#			 occurs)
#		-t	 set the title of the window to `title'
#		-o	 send output to `output_file'.  `-' means no file
#		-f	 read forms from `file'
#		form use `form' to send to lisp
#		...	 other arguments passed to lisp
#
# `-f file | form' must be just before `...'.  -t is ignore on UNIX.
# An output_file of `-' does not work on Windows.

# For debugging uncomment the following line:
#set -x

from_file=
output=
title=
batch="-batch -backtrace-on-error"
ignore_exit_status=xxx
plus_args="+M +B +n"

while test $# -gt 0; do
	case $1 in
	+p)	plus_args="$plus_args +p"
		;;
	-e)	ignore_exit_status=
		;;
	-f)	shift
		from_file=xxx
		form=$1
		;;
	-i)	batch=
		;;
	-o)	shift
		if test "X$1" != "X-"; then
			output=$1
		fi
		;;
	-t)	shift
		title="$1"
		;;
	*)	break
		;;
	esac
	shift
done

if test -z "$form"; then
	form=$1
	shift
fi

if test -n "$title" -a '(' -d c:/ -o -d //C/ ')'; then
	cl="$1 +t $title"
else
	cl=$1
fi
shift

other_lisp_args="$* $batch"

if test -d c:/ -o -d //C/; then # on Windows
# On Windows, we must have an output file, so send it to runlisp.out
# if the user didn't request one.
	if test -z "$output"; then
		output="runlisp.out"
	fi
	other_lisp_args="$other_lisp_args -d $output"
	if test -n "$from_file"; then
		$cl $plus_args +s $form $other_lisp_args
		status=$?
	else
		rm -f runlisp.tmp
		echo $form > runlisp.tmp
		$cl $plus_args +s runlisp.tmp $other_lisp_args
		status=$?
		rm -f runlisp.tmp
	fi
	if test -n "$ignore_exit_status" -a $status -ne 0 -a $status -ne 123; then
		echo ""
		echo "==== check \"$output\" for errors and/or warnings ===="
	fi
else # on UNIX
	if test -n "$from_file"; then
		start="cat $form"
	else
		start="echo $form"
	fi
	if test -n "$output"; then
		$start | $cl $other_lisp_args > $output 2>&1
	else
		$start | $cl $other_lisp_args
	fi
	status=$?
fi
exit $status
