#! /bin/sh
# $Id: update.sh,v 1.4.2.5 2001/12/12 01:37:53 layer Exp $
#
# Run ./update.sh to rebuild all our images to include patches.

do_bu=1

images="alisp.dxl alisp8.dxl mlisp.dxl mlisp8.dxl composer.dxl clim.dxl"

for image in $images; do
	if test ! -f $image; then
		continue
	fi

	base="`echo ${image} | sed 's/.dxl//'`"
	lisp=$base

	if test $do_bu -ne 0; then
	    do_bu=0
	    echo "Doing bundle check.  Output going to tmpbu.build..."
	    echo "The bundle check may take take several minutes."
	    cat << EOF | \
	    env ACL_LOCALE=C ./$lisp -I "$base".dxl -qq -batch -backtrace-on-error 2>&1 > tmpbu.build
(build-lisp-image "tmpbu.dxl" :verbose t :include-ide nil :include-devel-env t :internal-debug "tmpbu.debug")
(exit 0)
EOF
	    cat << EOF | \
	    env ACL_LOCALE=C ./$lisp -I tmpbu.dxl -qq -batch -backtrace-on-error 2>&1 >> tmpbu.build
(when (fboundp 'excl::update-bundle-check)
  (unless (excl::update-bundle-check t)
    (excl::update-bundle-files)))
(exit 0)
EOF
	    rm -f tmpbu.dxl
	    echo "Finished bundle check."
	fi

	if test ! -f $lisp; then
		echo "Sorry, $lisp does not exist."
		exit 1
	fi

	orig_image="$base"orig.dxl
	old_image="$base"old.dxl

	if test ! -f $orig_image; then
		mv $image $orig_image
		chmod a-w $orig_image
		prev=$orig_image
	fi
	if test -f $image; then
		mv $image $old_image
		prev=$old_image
	fi

	echo "Building $image..."
	cat << EOF | \
	env ACL_LOCALE=C ./$lisp -I $orig_image -qq -batch -backtrace-on-error 2>&1 > $base.build
(setq *print-startup-message*
  (let ((r (reverse *print-startup-message*)))
    (reverse
     (pushnew '(:external-format . t) r :test #'equal))))
(build-lisp-image "$image" :verbose t :internal-debug "$base.debug")
(exit 0)
EOF
	status=$?
	if test $status -ne 0; then
		echo "Building $image failed.  Will restore previous image."
		echo "Check $base.build for build errors."
		rm -f $image
		mv $prev $image
		exit 1
	fi
done
